// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#ifndef HPPWALKFOOTPLANNER_H
# define HPPWALKFOOTPLANNER_H
# include <KineoUtility/kitDefine.h>

# include <hpp/model/fwd.hh>
# include <hpp/core/planner.hh>

# include <hpp/walkfootplanner/fwd.hh>
# include <hpp/walkfootplanner/footstep-sequence.hh>
# include <hpp/walkfootplanner/config.hh>

namespace hpp
{
  namespace walkFootPlanner
  {
    /// \brief Collision-free foot step planning for humanoid robot.
    ///
    /// This class implements collision free foot step planning for a
    /// humanoid robot.  The robot is represented by a bounding box
    /// moving in the plane. Path planning is performed in two steps:
    ///
    /// \li first a collision free path is built for the bounding box,
    /// the default steering method is ChppWalkSteeringMethod,
    ///
    /// \li then sequences of step sequences (ChppWalkStepSequence)
    /// are computed along the path of the box.
    ///
    /// <b>How to use this class</b>
    ///
    /// To use this class,
    ///
    /// \li create an object Planner, by specifying the step
    /// size,
    /// \li create a robot bounding box with appropriate size
    /// (Planner::createRobotBox),
    /// \li insert this robot in the object (core::Planner::addHppProblem),
    /// \li call Planner::initializeProblem to set the default
    /// steering method and distance function,
    /// \li add obstacles using core::Planner::addObstacle,
    /// \li choose initial and goal configurations of the box using
    /// Planner::initConfig and Planner::goalConfig
    /// and
    /// \li call Planner::solve to compute a step sequences,
    /// \li the result can be accessed through
    /// Planner::sequenceOfStepSequences.
    class HPP_WALKFOOTPLANNER_DLLAPI Planner : public core::Planner
    {
    public:
      /// \brief Constructor
      ///
      /// \param inStepLength Abscissa of the landing foot in the frame of
      /// the support foot in straight line motions.
      /// \param inStepSpace Absolute value of the ordinate of the landing
      /// foot in the frame of the support foot
      /// \param inFootSecurityDistance Distance above which centers of
      /// feet should remain in two consecutive steps.
      /// \param addon whether the parser is embedded in and addon application.

      explicit Planner (double inStepLength,
			double inStepSpace,
			double inFootSecurityDistance,
			bool addon = true);

      virtual ~Planner ();

      /// \name Read access to parameters
      /// @{

      /// \brief Get step length
      double stepLength () const
      {
	return stepLength_;
      }

      /// \brief Get lateral step spacing
      double stepSpacing () const
      {
	return stepSpacing_;
      }

      /// \brief Get security distance between foot position on the ground.
      double footSecurityDistance() const
      {
	return footSecurityDistance_;
      }

      /// @}

      /// \name Problem resolution
      /// @{

      /// \brief Set initial configuration of Walk Planning problem
      /// \param x position and orientation of the robot.
      /// \param y position and orientation of the robot.
      /// \param theta position and orientation of the robot.
      virtual ktStatus initConfig (double x, double y, double theta);

      /// \brief Set goal configuration of Walk Planning problem
      /// \param x position and orientation of the robot.
      /// \param y position and orientation of the robot.
      /// \param theta position and orientation of the robot.
      virtual ktStatus goalConfig (double x, double y, double theta);

      /// \brief Create a robot the body of which is a box.
      ///
      /// \param xMin, xMax, yMin, yMax, zMin, zMax coordinates of the box
      /// vertices in joint frame.
      ///
      /// The root joint (and only one) is a freeflyer.  An extra dof is
      /// added to the robot in order to use CflicSteeringMethod.
      model::DeviceShPtr createRobotBox (double xMin, double yMin, double zMin,
				      double xMax, double yMax, double zMax);

      /// \brief Reset size of bounding box
      ///
      /// \param inXmin, inXmax, inYmin, inYmax, inZmin, inZmax coordinates
      /// of the box vertices in joint frame.
      bool setBoundingBoxSize (double inXmin, double inXmax, double inYmin,
			       double inYmax, double inZmin, double inZmax);

      /// \brief Initialize a Walk Planning problem.
      ///
      /// Create a steering method and distance function for walk planning.
      virtual ktStatus initializeProblem ();

      /// @}

      /// \name Problem resolution
      /// @{

      /// \brief Iteratively solve a walk motion planning problem for a
      /// humanoid robot.
      ///
      /// \return KD_OK or KD_ERROR
      virtual ktStatus solve ();

      /// \brief Compute the step sequence from the latest path of the box.
      ktStatus computeStepSequenceFromBoxPath ();

      /// \brief get the relativeFootPrint stack that follow a path
      /// \return attRelativeFootPrintVector
      ///
      /// The path of the bounding box is split into sections of
      /// discontinuous velocity.  A sequence of step is computed for each
      /// section. The robot is thus supposed to stop between two
      /// successive sequences.
      const FootStepSequence&
      footStepSequence () const
      {
	return footStepSequence_;
      }

      /// @}

    private:

      /// \brief compute from a box's path all the print foot for HRP2 :
      /// this is stored in m_RelativeFootPositions vector
      ///
      /// \param inBoxPath Path of the bounding box
      ///
      /// \retval outStepSequence sequence of steps
      void computeAStepSequenceFromAPath
      (CkwsPathShPtr inBoxPath,
       FootStepSequence& outStepSequence);

      /// \brief Get the robot bounding box
      CkppKCDBoxShPtr getRobotBoundingBox ();

      /// \brief Store pointer to bounding box of robot.
      CkppKCDBoxShPtr boundingBox_;

      /// \brief Step sequence of relative foot use to compute the path
      ///
      /// Previously ChppWalkPatGen::attRelativeFootPrintVector
      FootStepSequence footStepSequence_;

      /// \brief Length of the steps
      ///
      /// Abscissa of the landing foot in the frame of the support foot in
      /// straight line motions.
      double stepLength_;

      /// \brief Space between feet
      ///
      /// Absolute value of the ordinate of the landing foot in the frame
      /// of the support foot
      double stepSpacing_;

      /// \brief Security distance between feet
      ///
      /// Distance above which centers of feet should remain in two
      /// consecutive steps.
      double footSecurityDistance_;
    };

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.

#endif //! HPPWALKFOOTPLANNER_H
