// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#ifndef HPPWALKFOOTPLANNER_SPLINE_DIRECTPATH_HH
# define HPPWALKFOOTPLANNER_SPLINE_DIRECTPATH_HH
# include <boost/function.hpp>
# include <KineoWorks2/kwsPath.h>
# include <kwsPlus/directPath/kwsPlusDirectPath.h>

# include <roboptim/core/function.hh>
# include <roboptim/trajectory/fwd.hh>
# include <roboptim/trajectory/cubic-b-spline.hh>

# include <hpp/walkfootplanner/config.hh>
# include <hpp/walkfootplanner/fwd.hh>

namespace hpp
{
  namespace walkFootPlanner
  {
    /// Convert a Spline evaluation to a CkwsConfig.
    void convertResult (const roboptim::Function::result_t& val,
			const CkwsDeviceShPtr& inDevice,
			CkwsConfig& cfg);

    /// Scale t for free time trajectory.
    inline double scaleT (const roboptim::CubicBSpline& src,
			  const roboptim::CubicBSpline& dst, double t)
    {
      const double timeScale =
	dst.timeRange ().second / src.timeRange ().second;
      const double min = src.timeRange ().first;
      const double scaled = min + timeScale * (t - min);
      return scaled;
    }

    //FIXME: check that direct path are linear.
    /// \brief Default control point fitting method.
    ///
    /// This method fits the input path on the Spline.
    /// This approach uses N control points per direct path.
    ///
    /// For each direct path, the first N-1 are distributed in an
    /// uniform way.  The last one is at the same point than the
    /// previous one in order to make sure that the Spline is going
    /// through the control point (i.e. force a null speed).
    ///
    /// I.e. with N = 5 and three direct paths, control points will be:
    ///
    /// |||--|--|--||| |---|---|---||| |-|-|-|||
    ///
    /// ``|'' is a control point, ``-'' symbolizes the path, a space (`` '')
    /// is added to delimit each direct path.
    HPP_WALKFOOTPLANNER_DLLAPI roboptim::CubicBSpline::vector_t
    defaultControlPointsFittingStrategy (const CkwsPathShPtr& inPath);


    class HPP_WALKFOOTPLANNER_DLLAPI SplineDirectPath : public CkwsPlusDirectPath
    {
    public:
      static const int cfgSize = 3;
      typedef boost::function
	<roboptim::CubicBSpline::vector_t (const CkwsPathShPtr&)>
	controlPointsFittingStrategy_t;

      static double getX (const CkwsConfig&);
      static double getY (const CkwsConfig&);
      static double getTheta (const CkwsConfig&);

      virtual ~SplineDirectPath();

      static SplineDirectPathShPtr create (const CkwsPathShPtr&,
					   controlPointsFittingStrategy_t strategy
					   = &defaultControlPointsFittingStrategy);
      static SplineDirectPathShPtr create (const roboptim::CubicBSpline&,
					   const CkwsDeviceShPtr& inDevice);

      static SplineDirectPathShPtr
	createCopy (const SplineDirectPathConstShPtr& inDirectPath);

      SplineDirectPath (const SplineDirectPath& dp);

      const roboptim::CubicBSpline& spline () const
      {
	return attSpline;
      }

      //FIXME: can we get rid of that?
      // Is that behavior wanted?
      roboptim::CubicBSpline& spline ()
      {
	return attSpline;
      }

      virtual CkwsAbstractPathShPtr clone () const;
      virtual void maxAbsoluteDerivative
      (double, double, std::vector<double, std::allocator<double> >&) const;
      virtual void interpolate (double, CkwsConfig&) const;
      virtual double computePrivateLength () const;


    protected:
      SplineDirectPath (const CkwsPathShPtr&,
			CkwsConfigShPtr,
			CkwsConfigShPtr,
			controlPointsFittingStrategy_t strategy
			= &defaultControlPointsFittingStrategy);

      SplineDirectPath (const roboptim::CubicBSpline&,
			const CkwsDeviceShPtr& inDevice,
			CkwsConfigShPtr,
			CkwsConfigShPtr);
      ktStatus init (const SplineDirectPathWkPtr& inWeakPtr);

    private:
      roboptim::CubicBSpline attSpline;
      SplineDirectPathWkPtr attWeakPtr;
    };
  } // end of namespace walkFootPlanner.
} // end of namespace hpp.


#endif //! HPPWALKFOOTPLANNER_SPLINE_DIRECTPATH_HH
