// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <vector>

#include <boost/foreach.hpp>

#include <KineoKCDModel/kppKCDBox.h>
#include <KineoModel/kppJointComponent.h>
#include <KineoModel/kppSMLinearComponent.h>
#include <KineoModel/kppSolidComponentRef.h>
#include <KineoWorks2/kwsJointDof.h>
#include <KineoWorks2/kwsJointFreeFlyer.h>

#include <kwsPlus/directPath/kwsPlusSMLinear.h>

#include <hpp/model/body.hh>
#include <hpp/model/freeflyer-joint.hh>
#include <hpp/model/device.hh>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/walkfootplanner/humanoid-box-builder.hh"

namespace hpp
{
  namespace walkFootPlanner
  {
    namespace
    {
      static bool
      buildFlatBox (const double& x, const double& y,
		    const double& cx, const double& cy,
		    const double& zMin, const double& zMax,
		    bool isLeft,
		    const std::vector<int>& vertexVector,
		    model::BodyShPtr& hppBody);

      static bool
      buildFlatBox (const double& x, const double& y,
		    const double& cx, const double& cy,
		    const double& zMin, const double& zMax,
		    bool isLeft,
		    const std::vector<int>& vertexVector,
		    model::BodyShPtr& hppBody)
      {
	hppDebugStatement (char const* lr = (isLeft) ? "left" : "right");

	CkppKCDPolyhedronShPtr flatBox =
	  CkppKCDPolyhedron::create (isLeft ? "left.Flat.Box" : "right.Flat.Box");

	if (!flatBox)
	  {
	    hppDout (error, "failed to create polyhedron (" << lr << ")");
	    return false;
	  }


	if (flatBox->addPoint (CkitPoint3 (x, y, zMin)) == KD_ERROR)
	  {
	    hppDout (error, "failed to add point ("
		     << lr << ", point = "
		     << x << ", " << y << ", " << zMin << ")");
	    return false;
	  }

	if (flatBox->addPoint (CkitPoint3 (x, y, zMax)) == KD_ERROR)
	  {
	    hppDout (error, "failed to add point ("
		     << lr << ", point = "
		     << x << ", " << y << ", " << zMax << ")");
	    return false;
	  }

	if (flatBox->addPoint(CkitPoint3(cx, cy, zMax)) == KD_ERROR)
	  {
	    hppDout (error, "failed to add point ("
		     << lr << ", point = "
		     << cx << ", " << cy << ", " << zMax << ")");
	    return false;
	  }

	if (flatBox->addPoint (CkitPoint3 (cx, cy, zMin)) == KD_ERROR)
	  {
	    hppDout (error, "failed to add point ("
		     << lr << ", point = "
		     << cx << ", " << cy << ", " << zMin << ")");
	    return false;
	  }

	hppDout (info, "add flatBox (" << lr << ")");
	flatBox->addPolygon(vertexVector);

	if (flatBox->makeCollisionEntity(CkcdObject::IMMEDIATE_BUILD)
	    == KD_ERROR)
	  {
	    hppDout (error, "failed to create collision entities (" << lr << ")");
	    return false;
	  }
	hppBody->addInnerObject
	  (CkppSolidComponentRef::create (flatBox), CkitMat4 (), true);
	return true;
      }
    }


    HumanoidBoxBuilder::HumanoidBoxBuilder ()
    {}

    HumanoidBoxBuilder::~HumanoidBoxBuilder ()
    {}

    std::pair<model::DeviceShPtr, CkppKCDBoxShPtr>
    HumanoidBoxBuilder::operator()
      (const std::vector<CkcdObjectShPtr>& obstacleList,
       double xMin, double yMin, double zMin,
       double xMax, double yMax, double zMax)
    {
      const double sx = std::fabs (xMax - xMin);
      const double sy = std::fabs (yMax - yMin);
      const double sz = std::fabs (zMax - zMin);

      const double x = (xMax + xMin) / 2.;
      const double y = (yMax + yMin) / 2.;
      const double z = (zMax + zMin) / 2.;

      model::DeviceShPtr hppDevice = model::Device::create("Humanoid Box");
      hppDout (info, "Steering method: "
	       << hppDevice->steeringMethodComponent ()->name ());
      CkppKCDBoxShPtr boundingBox = CkppKCDBox::create ("H.Box", sx, sy, sz);

      CkitMat4 absPos;
      boundingBox->diffuseColor (CkppColor (0.2f, 0.2f, 0.6f, 0.4f));
      absPos.translate (x, y, z);
      boundingBox->setAbsolutePosition (absPos);
      boundingBox->makeCollisionEntity (CkcdObject::IMMEDIATE_BUILD);

      hppDout (info, "create bounding box with parameters:"
	       << incindent << iendl
	       << "xMin = " << xMin << iendl
	       << "xMax = " << xMax << iendl
	       << "yMin = " << yMin << iendl
	       << "yMax = " << yMax << iendl
	       << "zMin = " << zMin << iendl
	       << "zMax = " << zMax << iendl
	       << "size = (" << sx << ", " << sy << ", "<< sz << ")" << iendl
	       << "position = (" << x << ", " << y << ", "<< z << ")"
	       << decindent);

      // Create plan freeflyer joint
      model::JointShPtr hppJoint =
	model::FreeflyerJoint::create ("Rootjoint.H.Box", CkitMat4 ());
      CkppJointComponentShPtr kppJoint = hppJoint->kppJoint ();

      // Lock non-planar dofs
      {
	static const int Z = 2;
	static const int Rx = 3;
	static const int Ry = 4;
	kppJoint->kwsJoint ()->dof (Z)->isLocked (true);
	kppJoint->kwsJoint ()->dof (Rx)->isLocked (true);
	kppJoint->kwsJoint ()->dof (Ry)->isLocked (true);
      }

      // Display path of root joint
      kppJoint->doesDisplayPath (true);

      // Create body and attach to joint
      model::BodyShPtr hppBody = model::Body::create (boundingBox->name ());
      kppJoint->kwsJoint()->setAttachedBody (hppBody);
      hppJoint->jrlJoint()->setLinkedBody
	(*(model::Device::objectFactory()->createBody()));

      // Add bounding box to body
      hppBody->addInnerObject
	(CkppSolidComponentRef::create (boundingBox), absPos, false);

      // Attach joint to device
      hppDevice->setRootJoint (hppJoint);
      // Add obstacles of the environment to bounding box.

      // Set objects of the environment as obstacles for the box.
      BOOST_FOREACH (CkcdObjectShPtr obstacle, obstacleList)
	hppBody->addOuterObject (obstacle);

      // Create two flat boxes inside bounding box for distance computations.
      double halfLength = (xMax - xMin) / 2.;
      double halfWidth = (yMax - yMin) / 2.;

      double xr = 0., yr = 0., xl = 0., yl = 0.;
      double xc = (xMin + xMax) / 2.;
      double yc = (yMin + yMax) / 2.;

      std::vector<int> vertexVector (4);
      vertexVector[0] = 0;
      vertexVector[1] = 1;
      vertexVector[2] = 2;
      vertexVector[3] = 3;

      if (halfLength < halfWidth)
	{
	  xl = xc;
	  xr = xc;
	  yl = yc + (halfWidth - halfLength);
	  yr = yc - (halfWidth - halfLength);
	}
      else
	{
	  yl = yc;
	  yr = yc;
	  xl = xc + (halfLength - halfWidth);
	  xr = xc - (halfLength - halfWidth);
	}

      // Build flat boxes.
      if (!buildFlatBox (xl, yl, xc, yc, zMin, zMax, "left",
			 vertexVector, hppBody))
	return std::make_pair (model::DeviceShPtr (), CkppKCDBoxShPtr());

      if (!buildFlatBox (xr, yr, xc, yc, zMin, zMax, "right",
			 vertexVector, hppBody))
	return std::make_pair (model::DeviceShPtr (), CkppKCDBoxShPtr ());

      hppDevice->initialize ();

      hppDout (info, "Displaying device: "
	       << incindent << iendl << *hppDevice << decindent);

      return std::make_pair (hppDevice, boundingBox);
    }

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
