// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <boost/format.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/assign/list_of.hpp>

#include <roboptim/core/finite-difference-gradient.hh>
#include <roboptim/core/visualization/gnuplot.hh>
#include <roboptim/core/visualization/gnuplot-commands.hh>

#include <roboptim/trajectory/free-time-trajectory.hh>
#include <roboptim/trajectory/frontal-speed.hh>
#include <roboptim/trajectory/fwd.hh>
#include <roboptim/trajectory/orthogonal-speed.hh>
#include <roboptim/trajectory/cubic-b-spline.hh>
#include <roboptim/trajectory/stable-point-state-function.hh>
#include <roboptim/trajectory/trajectory-cost.hh>

#include <roboptim/trajectory/visualization/trajectory.hh>
#include <roboptim/trajectory/visualization/speed.hh>

#include <hpp/model/device.hh>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/walkfootplanner/minimize-time.hh"
#include "hpp/walkfootplanner/robot-obstacle-distance.hh"

using namespace roboptim;
using namespace roboptim::visualization;
using namespace roboptim::visualization::gnuplot;

// Problem parameters.
const unsigned configurationSpaceSize = 3;
const unsigned nConstraintsPerCtrlPts = 5;

namespace hpp
{
  namespace walkFootPlanner
  {

    namespace
    {
      CubicBSpline::matrix_t costAMatrix (const CubicBSpline& spline)
      {
	CubicBSpline::matrix_t a (1, 1 + spline.parameters ().size ());
	a.clear ();
	a (0, 0) = -1.;
	return a;
      }

      CubicBSpline::vector_t costBVector ()
      {
	CubicBSpline::vector_t b (1);
	b.clear ();
	return b;
      }

      class SpeedConstraint : public DerivableFunction
      {
      public:
	SpeedConstraint (value_type stepSpace,
			 bool leftFoot,
			 const double& vMaxX,
			 const double& vMaxY,
			 const double& vMaxTheta) throw ()
	  : DerivableFunction
	    (3 * 2, 1,
	     (boost::format
	      ("(v^x_{foot}/v^x_{max})^2 + (v^y_{foot}/v^y_{max})^2"
	       " + (\\dot{theta}/v^{\theta}_{max})^2 %1%")
	      % (leftFoot ? "left foot" : "right foot")).str ()),
	    stepSpace_ (stepSpace),
	    vMaxX_ (vMaxX),
	    vMaxY_ (vMaxY),
	    vMaxTheta_ (vMaxTheta),
	    epsilon_ (leftFoot ? -1. : 1.)
	{}

	~SpeedConstraint () throw ()
	{}

      protected:
	value_type computeVf (const argument_t& arg) const throw ()
	{
	  const value_type& theta = arg[2];
	  const value_type& dx = arg[3 + 0];
	  const value_type& dy = arg[3 + 1];
	  const value_type& dtheta = arg[3 + 2];
	  const value_type offset = .5*vMaxX_;

	  return dx * std::cos (theta)
	    + dy * std::sin (theta)
	    + epsilon_ * dtheta * stepSpace_ * .5
	    - offset;
	}

	value_type computeVorth (const argument_t& arg) const throw ()
	{
	  const value_type& theta = arg[2];
	  const value_type& dx = arg[3 + 0];
	  const value_type& dy = arg[3 + 1];

	  return dy * std::cos (theta) - dx * std::sin (theta);
	}

	gradient_t
	gradVf (const argument_t& arg) const throw ()
	{
	  const value_type& theta = arg[2];
	  const value_type& dx = arg[3 + 0];
	  const value_type& dy = arg[3 + 1];
	  const value_type& dtheta = arg[3 + 2];

	  gradient_t grad (inputSize ());
	  grad.clear ();
	  // Components 0 and 1 are null.
	  grad[2] =
	    dy * dtheta * std::cos (theta)
	    - dx * dtheta * std::sin (theta);
	  grad[3] = std::cos (theta);
	  grad[4] = std::sin (theta);
	  grad[5] = epsilon_ * stepSpace_ * .5;
	  return grad;
	}

	gradient_t
	gradVorth (const argument_t& arg) const throw ()
	{
	  const value_type& theta = arg[2];
	  const value_type& dx = arg[3 + 0];
	  const value_type& dy = arg[3 + 1];
	  const value_type& dtheta = arg[3 + 2];

	  gradient_t grad (inputSize ());
	  grad.clear ();
	  // Components 0, 1 and 5 are null.
	  grad[2] = -1. * (dx * dtheta * std::cos (theta)
			   + dy * dtheta * std::sin (theta));
	  grad[3] = -std::sin (theta);
	  grad[4] = std::cos (theta);
	  return grad;
	}

	void impl_compute (result_t& res, const argument_t& arg) const throw ()
	{
	  const value_type vf = computeVf (arg);
	  const value_type vorth = computeVorth (arg);

	  const value_type& dtheta = arg[3 + 2];

	  res[0] = -1.
	    + (vf / vMaxX_) * (vf / vMaxX_)
	    + (vorth / vMaxY_) * (vorth / vMaxY_)
	    + (dtheta / vMaxTheta_) * (dtheta / vMaxTheta_);
	  res[0] /= 2.;
	}

	void impl_gradient (gradient_t& grad, const argument_t& arg, size_type i)
	  const throw ()
	{
	  FiniteDifferenceGradient<finiteDifferenceGradientPolicies::Simple>
	    fd (*this, 1e-6);
	  fd.gradient (grad, arg, i);

	  // const value_type vf = computeVf (arg);
	  // const value_type vorth = computeVorth (arg);
	  // grad = alpha_ * vf * gradVorth (arg) + beta_ * vorth *
	  // gradVorth (arg);
	}

      private:
	value_type stepSpace_;
	const value_type& vMaxX_;
	const value_type& vMaxY_;
	const value_type& vMaxTheta_;
	value_type epsilon_;
      };

    }

    bool
    isValid (const MinimizeTime::freeTime_t& freeTimeTraj,
	     const double stepSpace,
	     const unsigned nConstraints,
	     const double& vMaxX,
	     const double& vMaxY,
	     const double& vMaxTheta);

    bool
    isValid (const MinimizeTime::freeTime_t& freeTimeTraj,
	     const double stepSpace,
	     const unsigned nConstraints,
	     const double& vMaxX,
	     const double& vMaxY,
	     const double& vMaxTheta)
    {
      using namespace boost;

      for (unsigned i = 1; i < nConstraints - 1; ++i)
	{
	  const double t = (i + 1.) / (nConstraints + 1.);
	  assert (t > 0. && t < 1.);

	  shared_ptr<DerivableFunction> scLeft
	    (new SpeedConstraint (stepSpace, true, vMaxX, vMaxY, vMaxTheta));
	  shared_ptr<DerivableFunction> scRight
	    (new SpeedConstraint (stepSpace, false, vMaxX, vMaxY, vMaxTheta));

	  shared_ptr<DerivableFunction> constraintLeft
	    (new StablePointStateFunction<FreeTimeTrajectory<CubicBSpline> >
	     (freeTimeTraj, scLeft, t * tMax, 1));
	  shared_ptr<DerivableFunction> constraintRight
	    (new StablePointStateFunction<FreeTimeTrajectory<CubicBSpline> >
	     (freeTimeTraj, scRight, t * tMax, 1));

	  if ((*constraintLeft) (freeTimeTraj.parameters ())[0] > 0)
	    return false;
	  if ((*constraintRight) (freeTimeTraj.parameters ())[0] > 0)
	    return false;
	}
      return true;
    }


    MinimizeTime::MinimizeTime (const CubicBSpline& spline,
				model::BodyShPtr body,
				SplineDirectPathShPtr directPath,
				double minDistance,
				double stepSpace,
				const double& vMaxX,
				const double& vMaxY,
				const double& vMaxTheta)
      throw ()
      : freeTimeTraj_ (spline, 1.),
	cost_ (costAMatrix (spline), costBVector ()),
	problem_ (cost_),
	stepSpace_ (stepSpace)
    {
      using namespace boost;
      using namespace boost::assign;

      const unsigned nControlPoints =
	spline.parameters ().size () / configurationSpaceSize;
      hppDout (info, "Number of control points: " << nControlPoints);
      const unsigned paramSize = freeTimeTraj_.parameters ().size ();
      hppDout (info, "Parameter size: " << paramSize);

      // Specify starting point.
      problem_.startingPoint () = freeTimeTraj_.parameters ();

      // Define start and end boundary condition.
      assert (spline.parameters ().size () >= 6);

      // Add constraints.

      // Freeze start/end point.
      spline.freezeCurveStart (problem_, 1);
      spline.freezeCurveEnd (problem_, 1);

      const unsigned nbObstacles = body->nbDistPairs ();
      Function::interval_t vRange = Function::makeUpperInterval (0.);
      const unsigned nConstraints = nControlPoints * nConstraintsPerCtrlPts;
      hppDout (info, "Number of constraints: " << nConstraints);

      for (unsigned i = 1; i < nConstraints - 1; ++i)
	{
	  const double t = (i + 1.) / (nConstraints + 1.);
	  assert (t > 0. && t < 1.);

	  // Add speed constraints.
	  shared_ptr<DerivableFunction> scLeft
	    (new SpeedConstraint (stepSpace_, true, vMaxX, vMaxY, vMaxTheta));
	  shared_ptr<DerivableFunction>
	    scRight (new SpeedConstraint
		     (stepSpace_, false, vMaxX, vMaxY, vMaxTheta));

	  shared_ptr<DerivableFunction> constraintLeft
	    (new StablePointStateFunction<FreeTimeTrajectory<CubicBSpline> >
	     (freeTimeTraj_, scLeft, t * tMax, 1));
	  shared_ptr<DerivableFunction> constraintRight
	    (new StablePointStateFunction<FreeTimeTrajectory<CubicBSpline> >
	     (freeTimeTraj_, scRight, t * tMax, 1));

	  problem_.addConstraint (constraintLeft, vRange);
	  problem_.addConstraint (constraintRight, vRange);

	  // Add obstacle constraints.
	  for (unsigned obstacle = 0; obstacle < nbObstacles; ++obstacle)
	    {
	      shared_ptr<DerivableFunction> robotObstacleDistance
		(new RobotObstacleDistance (directPath, t * tMax, obstacle));
	      problem_.addConstraint
		(robotObstacleDistance,
		 Function::makeLowerInterval (minDistance));
	    }
	}

      // Determine initial scale.
      Function::vector_t params = freeTimeTraj_.parameters ();
      while (!isValid (freeTimeTraj_, stepSpace_, nConstraints,
		       vMaxX, vMaxY, vMaxTheta))
	{
	  params[0] /= 1.1;
	  freeTimeTraj_.setParameters (params);
	  if (params[0] < 1e-3)
	    {
	      hppDout (error, "failed to find initial time scale");
	      break;
	    }
	}
      problem_.startingPoint () = params;
    }

    Gnuplot
    MinimizeTime::display (SplineDirectPathShPtr path) const throw ()
    {
      Gnuplot gnuplot = Gnuplot::make_interactive_gnuplot ();
      try {
	gnuplot
	  << gnuplot::set ("multiplot layout 1, 2 title 'trajectory'")
	  << gnuplot::set ("grid")
	  << plot_xytheta (path->spline ())
	  << plot_speeds (path->spline ())
	  << unset ("multiplot");
      }
      catch (std::string& err)
	{
	  hppDout (error, "Failed to display the problem: " << err);
	  hppDout (error, "Optimized spline is: " << path->spline ());
	}
      return gnuplot;
    }

    MinimizeTime::~MinimizeTime () throw ()
    {}

    const MinimizeTime::problem_t&
    MinimizeTime::getProblem () const throw ()
    {
      return problem_;
    }

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
