// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <iostream>
#include <fstream>

#include <boost/foreach.hpp>

#include <KineoKCDModel/kppKCDBox.h>
#include <KineoModel/kppExtraDofComponent.h>
#include <KineoModel/kppJointComponent.h>
#include <KineoModel/kppSolidComponentRef.h>
#include <KineoWorks2/kwsJointDof.h>
#include <KineoWorks2/kwsJointFreeFlyer.h>

#include <kwsPlus/directPath/kwsPlusSMLinear.h>

#include <hpp/model/body.hh>
#include <hpp/model/joint.hh>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/util/timer.hh"
#include "hpp/walkfootplanner/footstep-sequence-builder.hh"
#include "hpp/walkfootplanner/humanoid-box-builder.hh"
#include "hpp/walkfootplanner/io.hh"
#include "hpp/walkfootplanner/planner.hh"
#include "hpp/walkfootplanner/roboptim-optimizer.hh"
#include "hpp/walkfootplanner/spline-directpath.hh"

static const int BOUNDING_BOX = 0;
static const double TOR_OBSTACLE = 0.015;

static unsigned int boxConfigSize = 6;
static const unsigned DOF_X = 0;
static const unsigned DOF_Y = 1;
static const unsigned DOF_RZ = 5;


namespace hpp
{
  namespace walkFootPlanner
  {

    Planner::Planner (double inStepLength,
		      double inStepSpace,
		      double inSecurityDistance,
		      bool addon)
      : core::Planner(addon),
	stepLength_ (inStepLength),
	stepSpacing_ (inStepSpace),
	footSecurityDistance_ (inSecurityDistance)
    {}

    Planner::~Planner ()
    {}


    namespace
    {
      static std::vector<double>
      getDofValues
      (const double& x, const double& y, const double& theta);

      static std::vector<double>
      getDofValues
      (const double& x, const double& y, const double& theta)
      {
	std::vector<double> res (boxConfigSize, 0.);
	res[DOF_X] = x;
	res[DOF_Y] = y;
	res[DOF_RZ] = theta;
	return res;
      }

      static CkwsConfigShPtr
      makeBoxCfg (Planner& planner,
		  double x, double y, double theta);

      static CkwsConfigShPtr
      makeBoxCfg (Planner& planner,
		  double x, double y, double theta)
      {
	if (planner.getNbHppProblems () < 1)
	  {
	    hppDout (error, "missing robot");
	    return CkwsConfigShPtr ();
	  }
	if (!planner.robotIthProblem (BOUNDING_BOX))
	  {
	    hppDout (error, "empty bounding box");
	    return CkwsConfigShPtr ();

	  }

	return CkwsConfig::create(planner.robotIthProblem (BOUNDING_BOX),
				  getDofValues (x, y, theta));
      }
    }

    ktStatus
    Planner::initConfig (double x,
			 double y,
			 double theta)
    {
      CkwsConfigShPtr box = makeBoxCfg (*this, x, y, theta);

      if (!box)
	{
	  hppDout (error, "failed to create the box configuration");
	  return KD_ERROR;
	}

      initConfIthProblem(BOUNDING_BOX, box);
      return KD_OK;
    }


    ktStatus
    Planner::goalConfig (double x,
			 double y,
			 double theta)
    {
      CkwsConfigShPtr box = makeBoxCfg (*this, x, y, theta);

      if (!box)
	{
	  hppDout (error, "failed to create the box configuration");
	  return KD_ERROR;
	}

      resetGoalConfIthProblem (BOUNDING_BOX);
      addGoalConfIthProblem(BOUNDING_BOX, box);
      return KD_OK;
    }


    ktStatus
    Planner::initializeProblem()
    {
      if (getNbHppProblems() < 1)
	{
	  hppDout (error, "missing robot");
	  return KD_ERROR;
	}

      hppDout (info, "Set obstacles tolerances to default values");
      BOOST_FOREACH (CkcdObjectShPtr obstacle, obstacleList ())
	obstacle->tolerance (TOR_OBSTACLE);

      hppDout (info, "Initialize path optimizer");
      double boxHeight = 0.;
      {
	double x = 0.;
	double y = 0.;
	double z = 0.;
	getRobotBoundingBox ()->getSize (x, y, z);
	boxHeight = std::min (x, y) / 2.;
      }

      // Sensible default speeds.
      // vMaxX computed from pattern generator parameters
      static const double vMaxX = 0.5;
      static const double vMaxY = 0.1;
      static const double vMaxTheta = 0.1;

      CkwsPathOptimizerShPtr trajOptimizer =
	RobOptimOptimizer::create
	(boxHeight, stepSpacing_,
	 vMaxX, vMaxY, vMaxTheta,
	 this, &defaultControlPointsFittingStrategy);

      if (pathOptimizerIthProblem (0, trajOptimizer) == KD_ERROR)
	{
	  hppDout (error, "failed to set the path optimizer");
	  return KD_ERROR;
	}

      hppProblem (BOUNDING_BOX)->alwaysOptimize (true);

      hppDout (info, "problem initialization success");
      return KD_OK;
    }


#ifdef HPP_DEBUG
    namespace
    {
      static model::BodyShPtr
      getRootJointAttachedBody (core::Planner& planner);

      static model::BodyShPtr
      getRootJointAttachedBody (core::Planner& planner)
      {
	return KIT_DYNAMIC_PTR_CAST
	  (model::Body, planner.robotIthProblem
	   (BOUNDING_BOX)->rootJoint ()->attachedBody ());
      }
    }
#endif //! HPP_DEBUG

    ktStatus
    Planner::solve ()
    {
      hppDout (info, "number of pairs in distance computation:"
	       << getRootJointAttachedBody (*this)->nbDistPairs ());

      hppDout (info, "_______________ PHASE 1 _______________");

      if (!robotIthProblem (BOUNDING_BOX))
	{
	  hppDout (error, "missing bounding box");
	  return KD_ERROR ;
	}

      if(!initConfIthProblem (BOUNDING_BOX))
	{
	  hppDout (error, "missing initial configuration");
	  return KD_ERROR ;
	}

      if(goalConfIthProblem (BOUNDING_BOX).size () == 0)
	{
	  hppDout (error, "missing goal configuration");
	  return KD_ERROR ;
	}

      hppStartBenchmark (BOUNDING_BOX_PROBLEM);
      ktStatus result = solveOneProblem (BOUNDING_BOX);
      hppStopBenchmark (BOUNDING_BOX_PROBLEM);
      hppDisplayBenchmark (BOUNDING_BOX_PROBLEM);
      if (result == KD_ERROR)
	return KD_ERROR;

      hppDout (info, "_______________ PHASE 2 _______________");

      return computeStepSequenceFromBoxPath ();
    }


    ktStatus
    Planner::computeStepSequenceFromBoxPath()
    {
      if(getNbHppProblems () < 1)
	{
	  hppDout (error, "missing robot");
	  return KD_ERROR ;
	}

      CkwsPathShPtr pathBox = getPath (0, getNbPaths (0) - 1);
      if (!pathBox)
	{
	  hppDout (error, "missing bounding box path");
	  return KD_ERROR;
	}

      assert (pathBox->countDirectPaths () == 1);
      computeAStepSequenceFromAPath (pathBox, footStepSequence_);

      hppDout (info, footStepSequence_);
      return KD_OK;
    }


    void
    ChppWalkFootPlanner::
    computeAStepSequenceFromAPath(CkwsPathShPtr path,
				  FootStepSequence& footStepSequence)
    {
      FootStepSequenceBuilder fssb;
      fssb (path, footStepSequence, *this);
    }

    model::DeviceShPtr
    Planner::createRobotBox
    (double xMin, double yMin, double zMin,
     double xMax, double yMax, double zMax)
    {
      HumanoidBoxBuilder hbb;
      std::pair<model::DeviceShPtr, CkppKCDBoxShPtr> res = hbb
	(obstacleList (), xMin, yMin, zMin, xMax, yMax, zMax);

      boundingBox_ = res.second;
      return res.first;
    }

    bool
    Planner::setBoundingBoxSize
    (double xMin, double xMax, double yMin,
     double yMax, double zMin, double zMax)
    {
      bool status = true;
      // Check parameters validity.
      if (xMin >= xMax)
	{
	  hppDout (error,
		   "xMin >= xMax (xMin = " << xMin
		   << ", xMax = " << xMax << ")");
	  status = false;
	}
      if (yMin >= yMax)
	{
	  hppDout (error,
		   "yMin >= yMax (yMin = " << yMin
		   << ", yMax = " << yMax << ")");
	  status = false;
	}
      if (zMin >= zMax)
	{
	  hppDout (error,
		   "zMin >= zMax (zMin = " << zMin
		   << ", zMax = " << zMax << ")");
	  status = false;
	}
      if (!status)
	return false;

      // Compute center and size of the box.
      const double xCenter = (xMax + xMin) / 2.;
      const double xLength = (xMax - xMin);

      const double yCenter = (yMax + yMin) / 2.;
      const double yLength = (yMax - yMin);

      const double zCenter = (zMax + zMin) / 2.;
      const double zLength = (zMax - zMin);

      hppDout (info,
	       "xCenter= " << xCenter << iendl <<
	       "yCenter= " << yCenter << iendl <<
	       "zCenter= " << zCenter << iendl <<
	       "xLength= " << xLength << iendl <<
	       "yLength= " << yLength << iendl <<
	       "zLength= " << zLength);

      CkppKCDBoxShPtr box = getRobotBoundingBox ();
      if (!box)
	{
	  hppDout (error, "mising bounding box");
	  return false;
	}

      box->setSize (xLength, yLength, zLength);

      CkitMat4 relBoxPosition;
      relBoxPosition.translate (xCenter, yCenter, zCenter);
      box->setRelativePosition (relBoxPosition);

#ifdef HPP_DEBUG
      CkitMat4 absolutePositionBox;
      box->getAbsolutePosition (absolutePositionBox);
      hppDout (info, "bounding box absolute position:" << absolutePositionBox);
#endif // HPP_DEBUG

      return true;
    }


    CkppKCDBoxShPtr
    Planner::getRobotBoundingBox()
    {
      return boundingBox_;
    }

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
