// Copyright (C) 2008, 2009 by Florent Lamiraux, Thomas Moulard, CNRS.
//
// This file is part of the hpp-walkfootplanner.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//
// See the COPYING file for more information.

#include <algorithm>
#include <KineoWorks2/kwsPath.h>
#include <KineoWorks2/kwsDirectPath.h>
#include <KineoWorks2/kwsDevice.h>
#include <kwsPlus/directPath/kwsPlusSteeringMethodFactory.h>

#include <boost/numeric/ublas/vector_proxy.hpp>

#include <roboptim/core/io.hh>
#include <roboptim/trajectory/cubic-b-spline.hh>

#include "hpp/util/debug.hh"
#include "hpp/util/indent.hh"
#include "hpp/walkfootplanner/spline-directpath.hh"

using namespace roboptim;

static unsigned int boxConfigSize = 6;

static const unsigned DOF_X = 0;
static const unsigned DOF_Y = 1;
static const unsigned DOF_Z = 2;
static const unsigned DOF_RX = 3;
static const unsigned DOF_RY = 4;
static const unsigned DOF_RZ = 5;


namespace hpp
{
  namespace walkFootPlanner
  {
    void convertResult (const Function::result_t& val,
			const CkwsDeviceShPtr& inDevice,
			CkwsConfig& cfg)
    {
      assert(val.size() == 3 && inDevice->countDofs () == boxConfigSize);
      static std::vector<double> dofValues (boxConfigSize, 0.);
      for (unsigned i = 0; i < boxConfigSize; ++i)
	dofValues[i] = 0.;
      dofValues[DOF_X] = val[0];
      dofValues[DOF_Y] = val[1];
      dofValues[DOF_RZ] = val[2];
      cfg.setDofValues (dofValues);
    }

    namespace
    {
      /// Convert a uBlas vector into a vector<double>.
      std::vector<double> uBlasToVector (const CubicBSpline::result_t& src)
      {
	std::vector<double> res (src.size ());
	for (unsigned i = 0; i < src.size (); ++i)
	  res[i] = src[i];
	return res;
      }

      double computeLength (const CkwsPathShPtr& inPath)
      {
	double res = 0.;
	for (unsigned i = 0; i < inPath->countDirectPaths (); ++i)
	  {
	    CkwsDirectPathConstShPtr directPath = inPath->directPath(i);
	    res += directPath->length ();
	  }
	return res;
      }



      CkwsConfigShPtr getStartCfg(const roboptim::CubicBSpline& inSpline,
				  const CkwsDeviceShPtr& inDevice)
      {
	CkwsConfigShPtr res (CkwsConfig::create (inDevice));
	convertResult (inSpline (inSpline.timeRange ().first),
		       inDevice,
		       *res);
	return res;
      }

      CkwsConfigShPtr getEndCfg(const roboptim::CubicBSpline& inSpline,
				const CkwsDeviceShPtr& inDevice)
      {
	// Compute last point.
	using namespace boost::numeric::ublas;
	Function::vector_t lastControlPoint (inSpline.outputSize ());

	lastControlPoint = 1. / 6. * subrange
	  (inSpline.parameters (),
	   inSpline.parameters ().size () - 3 * inSpline.outputSize (),
	   inSpline.parameters ().size () - 2 * inSpline.outputSize ());

	lastControlPoint += 2. / 3. * subrange
	  (inSpline.parameters (),
	   inSpline.parameters ().size () - 2 * inSpline.outputSize (),
	   inSpline.parameters ().size () - 1 * inSpline.outputSize ());

	lastControlPoint += 1. / 6. * subrange
	  (inSpline.parameters (),
	   inSpline.parameters ().size () - 1 * inSpline.outputSize (),
	   inSpline.parameters ().size () - 0 * inSpline.outputSize ());

	CkwsConfigShPtr res (CkwsConfig::create (inDevice));
	convertResult(lastControlPoint, inDevice, *res);
	return res;
      }

      CkwsConfigShPtr getStartCfg(const CkwsPathShPtr& inPath)
      {
	CkwsConfigShPtr cfg(new CkwsConfig(inPath->device()));
	inPath->getConfigAtStart(*cfg);
	return cfg;
      }

      CkwsConfigShPtr getEndCfg(const CkwsPathShPtr& inPath)
      {
	CkwsConfigShPtr cfg(new CkwsConfig(inPath->device()));
	inPath->getConfigAtEnd(*cfg);
	return cfg;
      }

    } // end of anonymous namespace.

    roboptim::CubicBSpline::vector_t
    defaultControlPointsFittingStrategy (const CkwsPathShPtr& inPath)
    {
      unsigned ctrlPointsPerDirectPath = 3;
      if (inPath->countDirectPaths() == 1)
	ctrlPointsPerDirectPath = 6;

      assert (ctrlPointsPerDirectPath >= 3);
      static const unsigned cfgSize = SplineDirectPath::cfgSize;

      static const unsigned paramSize =
	cfgSize * (ctrlPointsPerDirectPath * inPath->countDirectPaths () + 3);
      CubicBSpline::vector_t params (paramSize);

      hppDout (info, "parameters vector size: " << paramSize);

      // Create a parameters vector from the set of configurations
      // of the path.
      params.clear ();

      // Offset on angle
      double angleOffset = 0.;
      double thetaPrev = 0.;
      double theta = 0.;
      unsigned int indexInParam = 0;
      CkwsConfig cfg (inPath->device ());
      for (unsigned i = 0; i <= inPath->countDirectPaths (); ++i) {
	bool lastDirectPath = false;
	unsigned int ctrlPointThisDirectPath = ctrlPointsPerDirectPath;
	CkwsDirectPathConstShPtr directPath;
	if (i == inPath->countDirectPaths ()) {
	  lastDirectPath = true;
	  directPath = inPath->directPath(i-1);
	  ctrlPointThisDirectPath = 3;
	} else {
	  directPath = inPath->directPath(i);
	}
	for (unsigned n = 0; n < ctrlPointThisDirectPath; ++n) {
	  double t = 0.;

	  if (!lastDirectPath) {
	    // The first three points are at the beginning of the direct path.
	    if (n <= 2)
	      t = 0;
	    else
	      t = directPath->length()*((n-2)/(ctrlPointThisDirectPath-2.));
	  } else {
	    t = directPath->length();
	  }
	  directPath->getConfigAtDistance (t, cfg);
	  assert (cfg.size () == boxConfigSize);

	  params[indexInParam] = SplineDirectPath::getX (cfg);
	  indexInParam++;
	  params[indexInParam] = SplineDirectPath::getY (cfg);
	  indexInParam++;
	  theta = SplineDirectPath::getTheta (cfg) + angleOffset;

	  // Make theta continuous
	  bool offsetChanged = false;
	  if (theta - thetaPrev > M_PI)
	    theta -= M_PI * 2, angleOffset -= M_PI * 2, offsetChanged = true;
	  else if (theta - thetaPrev < -M_PI)
	    theta += M_PI * 2, angleOffset += M_PI * 2, offsetChanged = true;

	  if (offsetChanged)
	    hppDout (info, "offset: " << angleOffset << iendl
		     << "previous theta: " << thetaPrev << iendl
		     << "theta: " << theta);
	  thetaPrev = theta;

	  params[indexInParam] = theta;
	  indexInParam++;
	}
      }
      assert(paramSize == indexInParam);
      return params;
    }

    double
    SplineDirectPath::getX (const CkwsConfig& cfg)
    {
      return cfg.dofValue (DOF_X);
    }

    double
    SplineDirectPath::getY (const CkwsConfig& cfg)
    {
      return cfg.dofValue (DOF_Y);
    }

    double
    SplineDirectPath::getTheta (const CkwsConfig& cfg)
    {
      return cfg.dofValue (DOF_RZ);
    }

    SplineDirectPath::~SplineDirectPath()
    {
    }

    SplineDirectPathShPtr
    SplineDirectPath::create(const CkwsPathShPtr& inPath,
			     controlPointsFittingStrategy_t strategy)
    {
      CkwsConfigShPtr startCfg = getStartCfg (inPath);
      CkwsConfigShPtr endCfg = getEndCfg (inPath);

      SplineDirectPath* spline = new SplineDirectPath
	(inPath, startCfg, endCfg, strategy);
      SplineDirectPathShPtr splineShPtr(spline);
      SplineDirectPathWkPtr splineWkPtr(splineShPtr);

      if (spline->init(splineWkPtr) != KD_OK)
	splineShPtr.reset();
      return splineShPtr;
    }

    SplineDirectPathShPtr
    SplineDirectPath::create(const CubicBSpline& inSpline,
			     const CkwsDeviceShPtr& inDevice)
    {
      CkwsConfigShPtr startCfg = getStartCfg (inSpline, inDevice);
      CkwsConfigShPtr endCfg = getEndCfg (inSpline, inDevice);
      SplineDirectPath* spline = new SplineDirectPath
	(inSpline, inDevice, startCfg, endCfg);
      SplineDirectPathShPtr splineShPtr (spline);
      SplineDirectPathWkPtr splineWkPtr(splineShPtr);

      if (spline->init(splineWkPtr) != KD_OK)
	splineShPtr.reset();

      return splineShPtr;
    }

    SplineDirectPathShPtr
    SplineDirectPath::createCopy
    (const SplineDirectPathConstShPtr& inDirectPath)
    {
      if(inDirectPath) {
	SplineDirectPath* pathPtr =
	  new SplineDirectPath(*inDirectPath);
	SplineDirectPathShPtr pathShPtr(pathPtr);
	SplineDirectPathWkPtr pathWkPtr(pathShPtr);

	if(pathPtr->init(pathWkPtr) != KD_OK) {
	  pathShPtr.reset() ;
	}

	return pathShPtr;

      }
      else return SplineDirectPathShPtr() ;
    }

    SplineDirectPath::SplineDirectPath
    (const CkwsPathShPtr& inPath,
     CkwsConfigShPtr startCfg,
     CkwsConfigShPtr endCfg,
     controlPointsFittingStrategy_t strategy)
      : CkwsPlusDirectPath (*startCfg, *endCfg,
			    inPath->device ()->steeringMethod ()),
	attSpline (Function::makeInterval (0., computeLength (inPath)),
		   cfgSize,
		   strategy (inPath)),
	attWeakPtr ()
    {
    }

    SplineDirectPath::SplineDirectPath(const CubicBSpline& inSpline,
				       const CkwsDeviceShPtr& inDevice,
				       CkwsConfigShPtr startCfg,
				       CkwsConfigShPtr endCfg)
      : CkwsPlusDirectPath (*startCfg, *endCfg, inDevice->steeringMethod ()),
	attSpline (inSpline),
	attWeakPtr ()
    {
    }

    SplineDirectPath::SplineDirectPath(const SplineDirectPath& dp)
      : CkwsPlusDirectPath (dp),
	attSpline (dp.attSpline),
	attWeakPtr ()
    {
    }

    ktStatus
    SplineDirectPath::init(const SplineDirectPathWkPtr& inWeakPtr)
    {
      if (CkwsDirectPath::init(inWeakPtr) != KD_OK)
	return KD_ERROR;
      attWeakPtr = inWeakPtr;
      return KD_OK;
    }

    double
    SplineDirectPath::computePrivateLength() const
    {
      return attSpline.length ();
    }

    void
    SplineDirectPath::interpolate(double i_s, CkwsConfig& o_cfg) const
    {
      assert(i_s <= privateLength());

      CubicBSpline::vector_t result =
	attSpline(std::min(i_s, privateLength ()));
      for(unsigned i = 0; i < o_cfg.size(); ++i)
	o_cfg.dofValue(i, 0.);
      o_cfg.dofValue(DOF_X, result[0]);
      o_cfg.dofValue(DOF_Y, result[1]);
      o_cfg.dofValue(DOF_RZ, result[2]);
    }

    void
    SplineDirectPath::maxAbsoluteDerivative(double i_from,
					    double i_to,
					    std::vector<double>& o_derivative)
      const
    {
      for (unsigned i = 0; i < o_derivative.size (); ++i)
	o_derivative[i] = 0.;

      for (double t = i_from; t < i_to; t += .1)
	{
	  if (t > i_to)
	    return;
	  CubicBSpline::gradient_t gradient = attSpline.derivative (t, 1);
	  for (unsigned i = 0; i < o_derivative.size (); ++i)
	    o_derivative[i] = std::max (o_derivative[i], fabs (gradient[i]));
	}
    }

    CkwsAbstractPathShPtr
    SplineDirectPath::clone () const
    {
      return SplineDirectPath::createCopy(attWeakPtr.lock());
    }

  } // end of namespace walkFootPlanner.
} // end of namespace hpp.
